/*
 * FreeRTOS Kernel V10.4.3
 * Copyright (C) 2020 Amazon.com, Inc. or its affiliates.  All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * https://www.FreeRTOS.org
 * https://github.com/FreeRTOS
 *
 * 1 tab == 4 spaces!
 */

	.text
	.arm

	.set SYS_MODE,	0x1f
	.set SVC_MODE,	0x13
	.set IRQ_MODE,	0x12

	/* Hardware registers. */
	.extern ulICCIAR
	.extern ulICCEOIR
	.extern ulICCPMR

	/* Variables and functions. */
	.extern ulMaxAPIPriorityMask
	.extern _freertos_vector_table
	.extern pxCurrentTCB
	.extern vTaskSwitchContext
	.extern vApplicationIRQHandler
	.extern ulPortInterruptNesting
	.extern ulPortTaskHasFPUContext

	.global FreeRTOS_IRQ_Handler
	.global FreeRTOS_SWI_Handler
	.global vPortRestoreTaskContext
	.global portStartFirstTask

	.equ UNLOCKED, 0xFF

.macro portSAVE_CONTEXT

	/* Save the LR and SPSR onto the system mode stack before switching to
	system mode to save the remaining system mode registers. */
	SRSDB	sp!, #SYS_MODE
	CPS		#SYS_MODE
	PUSH	{R0-R12, R14}

	/* Does the task have a floating point context that needs saving?  If
	ulPortTaskHasFPUContext is 0 then no. */
	LDR		R2, ulPortTaskHasFPUContextConst
	LDR		R3, [R2]
	CMP		R3, #0

	/* Save ulPortTaskHasFPUContext itself. */
	PUSH	{R3}

	/* Save the stack pointer in the TCB. */
	MRC     p15, 0, R2, C0, C0, 5           /* Read CPU ID register*/
    AND     R2, R2, #0x03                   /* Mask off, leaving the CPU ID field*/
	LDR		R0, pxCurrentTCBConst			/* R0 = pxCurrentTCBs[0] */
	LSLS    R2, #2
	LDR		R0, [R0, R2]					/* R0 = pxCurrentTCBs[CPI_ID] */
	STR		SP, [R0]

	.endm

; /**********************************************************************/

.macro portRESTORE_CONTEXT

	/* Set the SP to point to the stack of the task being restored. */
	MRC     p15, 0, R2, C0, C0, 5           /* Read CPU ID register*/
    AND     R2, R2, #0x03                   /* Mask off, leaving the CPU ID field*/
	LDR		R0, pxCurrentTCBConst			/* R0 = pxCurrentTCBs[0] */
	LSLS    R2, #2
	LDR		R0, [R0, R2]					/* R0 = pxCurrentTCBs[CPI_ID] */
	LDR		SP, [R0]

	/* Is there a floating point context to restore?  If the restored
	ulPortTaskHasFPUContext is zero then no. */
	LDR		R0, ulPortTaskHasFPUContextConst
	POP		{R1}
	STR		R1, [R0]
	CMP		R1, #0

	/* Restore all system mode registers other than the SP (which is already
	being used). */
	POP		{R0-R12, R14}

	/* Return to the task code, loading CPSR on the way. */
	RFEIA	sp!

	.endm

; /**********************************************************************/
.macro portStartFirstTask
	MRC     p15, 0, R2, C0, C0, 5           /* Read CPU ID register*/
    AND     R2, R2, #0x03                   /* Mask off, leaving the CPU ID field*/
	LDR		R1, pxCurrentTCBConst			/* R1 = pxCurrentTCBs[0] */
	LDR		R1, [R1, R2]					/* R1 = pxCurrentTCBs[CPI_ID] */
	LDR  	R0, [R1]						/* The first item in pxCurrentTCB is the task top of stack. */
	ADDS 	R0, #32							/* Discard everything up to r0. */

	.endm




/******************************************************************************
 * SVC handler is used to start the scheduler.
 *****************************************************************************/
.align 4
.type FreeRTOS_SWI_Handler, %function
FreeRTOS_SWI_Handler:
	/* Save the context of the current task and select a new task to run. */
	portSAVE_CONTEXT
	MRC     p15, 0, R0, C0, C0, 5           /* Read CPU ID register*/
    AND     R0, R0, #0x03                   /* Mask off, leaving the CPU ID field*/
	CPSID	i
	BL vTaskSwitchContext
	CPSIE	i
	portRESTORE_CONTEXT


/******************************************************************************
 * vPortRestoreTaskContext is used to start the scheduler.
 *****************************************************************************/
.type vPortRestoreTaskContext, %function
vPortRestoreTaskContext:
	/* Switch to system mode. */
	CPS		#SYS_MODE
	portRESTORE_CONTEXT

/******************************************************************************
 * vportStartFirstTask is used to start the scheduler.
 *****************************************************************************/
.type vportStartFirstTask, %function
vportStartFirstTask:
	/* Switch to system mode. */
	CPS		#SYS_MODE
	portStartFirstTask

.align 4
.type FreeRTOS_IRQ_Handler, %function
FreeRTOS_IRQ_Handler:

	/* Return to the interrupted instruction. */
	SUB		lr, lr, #4

	/* Push the return address and SPSR. */
	PUSH	{lr}
	MRS		lr, SPSR
	PUSH	{lr}

	/* Change to supervisor mode to allow reentry. */
	CPS		#SVC_MODE

	/* Push used registers. */
	PUSH	{r0-r4, r12}

	/* Increment nesting count.  r3 holds the address of ulPortInterruptNesting
	for future use.  r1 holds the original ulPortInterruptNesting value for
	future use. */
	LDR		r3, ulPortInterruptNestingConst
	LDR		r1, [r3]
	ADD		r4, r1, #1
	STR		r4, [r3]

	/* Read value from the interrupt acknowledge register, which is stored in r0
	for future parameter and interrupt clearing use. */
	LDR 	r2, ulICCIARConst
	LDR		r2, [r2]
	LDR		r0, [r2]

	/* Ensure bit 2 of the stack pointer is clear.  r2 holds the bit 2 value for
	future use.  _RB_ Is this ever needed provided the start of the stack is
	alligned on an 8-byte boundary? */
	MOV		r2, sp
	AND		r2, r2, #4
	SUB		sp, sp, r2

	/* Call the interrupt handler. */
	PUSH	{r0-r4, lr}
	LDR		r1, vApplicationIRQHandlerConst
	BLX		r1
	POP		{r0-r4, lr}
	ADD		sp, sp, r2

	CPSID	i
	DSB
	ISB

	/* Write the value read from ICCIAR to ICCEOIR. */
	LDR 	r4, ulICCEOIRConst
	LDR		r4, [r4]
	STR		r0, [r4]

	/* Restore the old nesting count. */
	STR		r1, [r3]

	/* A context switch is never performed if the nesting count is not 0. */
	CMP		r1, #0
	BNE		exit_without_switch

	/* Did the interrupt request a context switch?  r1 holds the address of
	ulPortYieldRequired and r0 the value of ulPortYieldRequired for future
	use. */
	MRC     p15, 0, r1, C0, C0, 5           /* Read CPU ID register*/
    AND     r1, r1, #0x03                   /* Mask off, leaving the CPU ID field*/
	LDR		r2, =ulPortYieldRequired		/* r2 = ulPortYieldRequired[0]*/
	LDR		r0, [r1, r2]				    /* r0 = ulPortYieldRequired[CPU_ID]*/
	CMP		r0, #0
	BNE		switch_before_exit

exit_without_switch:
	/* No context switch.  Restore used registers, LR_irq and SPSR before
	returning. */
	POP		{r0-r4, r12}
	CPS		#IRQ_MODE
	POP		{LR}
	MSR		SPSR_cxsf, LR
	POP		{LR}
	MOVS	PC, LR

switch_before_exit:
	/* A context swtich is to be performed.  Clear the context switch pending
	flag. */
	MOV		r0, #0
	STR		r0, [r1]

	/* Restore used registers, LR-irq and SPSR before saving the context
	to the task stack. */
	POP		{r0-r4, r12}
	CPS		#IRQ_MODE
	POP		{LR}
	MSR		SPSR_cxsf, LR
	POP		{LR}
	portSAVE_CONTEXT

	/* Call the function that selects the new task to execute.
	vTaskSwitchContext() if vTaskSwitchContext() uses LDRD or STRD
	instructions, or 8 byte aligned stack allocated data.  LR does not need
	saving as a new LR will be loaded by portRESTORE_CONTEXT anyway. */
	MRC     p15, 0, R0, C0, C0, 5           /* Read CPU ID register*/
    AND     R0, R0, #0x03                   /* Mask off, leaving the CPU ID field*/
	CPSID	i
	BL vTaskSwitchContext
	CPSIE	i

	/* Restore the context of, and branch to, the task selected to execute
	next. */
	portRESTORE_CONTEXT


// ------------------------------------------------------------

  .global lockMutex
  // void lockMutex(mutex_t* pMutex)
  // Blocking call, returns once successfully locked a mutex
  // r0 = address of mutex_t
  .type lockMutex, "function"
  .cfi_startproc
lockMutex:

  // Is mutex locked?
  // -----------------
  LDREX   r1, [r0]                // Read lock field
  CMP     r1, #UNLOCKED           // Compare with "unlocked"

  //WFENE                           // If mutex is locked, go into standby
  BNE     lockMutex               // On waking re-check the mutex

  // Attempt to lock mutex
  // -----------------------
  MRC     p15, 0, r1, c0, c0, 5   // Read CPU ID register
  AND     r1, r1, #0x03           // Mask off, leaving the CPU ID field.
  STREX   r2, r1, [r0]            // Attempt to lock mutex, by write CPU's ID to lock field
  CMP     r2, #0x0                // Check whether store completed successfully (0=succeeded)
  BNE     lockMutex               // If store failed, go back to beginning and try again

  DMB

  BX      lr                      // Return as mutex is now locked by this cpu
  .cfi_endproc


// ------------------------------------------------------------

  .global unlockMutex
  // unsigned int unlockMutex(mutex_t* pMutex)
  // Releases mutex, returns 0x0 for success and 0x1 for failure
  // r0 = address of mutex_t
  .type unlockMutex, "function"
  .cfi_startproc
unlockMutex:

  // Does this CPU own the mutex?
  // -----------------------------
  MRC     p15, 0, r1, c0, c0, 5   // Read CPU ID register
  AND     r1, r1, #0x03           // Mask off, leaving the CPU ID in r1
  LDR     r2, [r0]                // Read the lock field of the mutex
  CMP     r1, r2                  // Compare ID of this CPU with the lock owner
  MOVNE   r0, #0x1                // If ID doesn't match, return "fail"
  BXNE    lr


  // Unlock mutex
  // -------------
  DMB                             // Ensure that accesses to shared resource have completed

  MOV     r1, #UNLOCKED           // Write "unlocked" into lock field
  STR     r1, [r0]

  DSB                             // Ensure that no instructions following the barrier execute until
                                  // all memory accesses prior to the barrier have completed.

  SEV                             // Send event to other CPUs, wakes anyone waiting on a mutex (using WFE)

  MOV     r0, #0x0                // Return "success"
  BX      lr
  .cfi_endproc


ulICCIARConst:	.word ulICCIAR
ulICCEOIRConst:	.word ulICCEOIR
ulICCPMRConst: .word ulICCPMR
pxCurrentTCBConst: .word pxCurrentTCBs
ulCriticalNestingConst: .word ulCriticalNesting
ulPortTaskHasFPUContextConst: .word ulPortTaskHasFPUContext
ulMaxAPIPriorityMaskConst: .word ulMaxAPIPriorityMask
vTaskSwitchContextConst: .word vTaskSwitchContext
vApplicationIRQHandlerConst: .word vApplicationIRQHandler
ulPortInterruptNestingConst: .word ulPortInterruptNesting

.end





