/*
 * FreeRTOS Kernel V10.4.3
 * Copyright (C) 2020 Amazon.com, Inc. or its affiliates.  All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * https://www.FreeRTOS.org
 * https://github.com/FreeRTOS
 *
 * 1 tab == 4 spaces!
 */

#ifndef PORTMACRO_H
    #define PORTMACRO_H

    #ifdef __cplusplus
        extern "C" {
    #endif

/*-----------------------------------------------------------
 * Port specific definitions.
 *
 * The settings in this file configure FreeRTOS correctly for the given hardware
 * and compiler.
 *
 * These settings should not be altered.
 *-----------------------------------------------------------
 */

/* Type definitions. */
    #define portCHAR          char
    #define portFLOAT         float
    #define portDOUBLE        double
    #define portLONG          long
    #define portSHORT         short
    #define portSTACK_TYPE    uint32_t
    #define portBASE_TYPE     long

    typedef portSTACK_TYPE   StackType_t;
    typedef long             BaseType_t;
    typedef unsigned long    UBaseType_t;

    typedef uint32_t         TickType_t;
    #define portMAX_DELAY    ( TickType_t ) 0xffffffffUL

/*-----------------------------------------------------------*/

/* Hardware specifics. */
    #define portSTACK_GROWTH      ( -1 )
    #define portTICK_PERIOD_MS    ( ( TickType_t ) 1000 / configTICK_RATE_HZ )
    #define portBYTE_ALIGNMENT    8

/*-----------------------------------------------------------*/

/* Task utilities. */

/* Called at the end of an ISR that can cause a context switch. */
    #define portEND_SWITCHING_ISR( xSwitchRequired ) \
    do \
    { \
        if( xSwitchRequired != pdFALSE ) \
        { \
            extern uint32_t ulPortYieldRequired[ portMAX_CORE_COUNT ]; \
            ulPortYieldRequired[ portGET_CORE_ID() ] = pdTRUE; \
        } \
    } while( 0 )

    #define portYIELD_FROM_ISR( x )    portEND_SWITCHING_ISR( x )
    #define portYIELD()                __asm volatile ( "SWI 0" ::: "memory" );
    //#define portYIELD()                 vYield()


/*-----------------------------------------------------------
* Critical section control
*----------------------------------------------------------*/

    extern void vPortEnterCritical( void );
    extern void vPortExitCritical( void );
    extern uint32_t ulPortSetInterruptMask( void );
    extern void vPortClearInterruptMask( uint32_t ulNewMaskValue );
    extern void vPortInstallFreeRTOSVectorTable( void );

/* These macros do not globally disable/enable interrupts.  They do mask off
 * interrupts that have a priority below configMAX_API_CALL_INTERRUPT_PRIORITY. */
/*
 * Save the current interrupt enable state & disable IRQs
 */
#define portDISABLE_INTERRUPTS()({                           \
       uint32_t ulState;                                     \
       __asm volatile ("mrs %0, cpsr" : "=r" (ulState)::);   \
       __asm volatile ( " cpsid i " ::: "memory" );          \
       ulState;})
/*
 * restore saved IRQ & FIQ state
 */
#define portRESTORE_INTERRUPTS(x)                           \
    __asm__ __volatile__(                                   \
    "msr    cpsr_c, %0              @ local_irq_restore\n"  \
    :                                                       \
    : "r" (x)                                               \
    : "memory", "cc")
    
    extern void vTaskEnterCritical( void );
    extern void vTaskExitCritical( void );
    #define portENTER_CRITICAL()                      vTaskEnterCritical();
    #define portEXIT_CRITICAL()                       vTaskExitCritical();
    #define portENABLE_INTERRUPTS()                   vPortClearInterruptMask( 0 )
    #define portSET_INTERRUPT_MASK_FROM_ISR()         ({ uint32_t ulStateISR = portDISABLE_INTERRUPTS(); vTaskEnterCritical(); ulStateISR; })
    #define portCLEAR_INTERRUPT_MASK_FROM_ISR( x )    do { vTaskExitCritical(); portRESTORE_INTERRUPTS( x ); } while (0)

/*-----------------------------------------------------------*/

/* Task function macros as described on the FreeRTOS.org WEB site.  These are
 * not required for this port but included in case common demo code that uses these
 * macros is used. */
    #define portTASK_FUNCTION_PROTO( vFunction, pvParameters )    void vFunction( void * pvParameters )
    #define portTASK_FUNCTION( vFunction, pvParameters )          void vFunction( void * pvParameters )

/* Prototype of the FreeRTOS tick handler.  This must be installed as the
 * handler for whichever peripheral is used to generate the RTOS tick. */
    void FreeRTOS_Tick_Handler( void );

/* Any task that uses the floating point unit MUST call vPortTaskUsesFPU()
 * before any floating point instructions are executed. */
    void vPortTaskUsesFPU( void );
    #define portTASK_USES_FLOATING_POINT()    vPortTaskUsesFPU()

    #define portLOWEST_INTERRUPT_PRIORITY           ( ( ( uint32_t ) configUNIQUE_INTERRUPT_PRIORITIES ) - 1UL )
    #define portLOWEST_USABLE_INTERRUPT_PRIORITY    ( portLOWEST_INTERRUPT_PRIORITY - 1UL )

/* Architecture specific optimisations. */
    #ifndef configUSE_PORT_OPTIMISED_TASK_SELECTION
        #define configUSE_PORT_OPTIMISED_TASK_SELECTION    1
    #endif

    #if configUSE_PORT_OPTIMISED_TASK_SELECTION == 1

/* Store/clear the ready priorities in a bit map. */
        #define portRECORD_READY_PRIORITY( uxPriority, uxReadyPriorities )    ( uxReadyPriorities ) |= ( 1UL << ( uxPriority ) )
        #define portRESET_READY_PRIORITY( uxPriority, uxReadyPriorities )     ( uxReadyPriorities ) &= ~( 1UL << ( uxPriority ) )

/*-----------------------------------------------------------*/

        #define portGET_HIGHEST_PRIORITY( uxTopPriority, uxReadyPriorities )    uxTopPriority = ( 31 - __builtin_clz( uxReadyPriorities ) )

    #endif /* configUSE_PORT_OPTIMISED_TASK_SELECTION */

    #ifdef configASSERT
        void vPortValidateInterruptPriority( void );
        #define portASSERT_IF_INTERRUPT_PRIORITY_INVALID()    vPortValidateInterruptPriority()
    #endif /* configASSERT */

    #define portNOP()                                         __asm volatile ( "NOP" )


    #ifdef __cplusplus
        } /* extern C */
    #endif


/* The number of bits to shift for an interrupt priority is dependent on the
 * number of bits implemented by the interrupt controller. */
    #if configUNIQUE_INTERRUPT_PRIORITIES == 16
        #define portPRIORITY_SHIFT            4
        #define portMAX_BINARY_POINT_VALUE    3
    #elif configUNIQUE_INTERRUPT_PRIORITIES == 32
        #define portPRIORITY_SHIFT            3
        #define portMAX_BINARY_POINT_VALUE    2
    #elif configUNIQUE_INTERRUPT_PRIORITIES == 64
        #define portPRIORITY_SHIFT            2
        #define portMAX_BINARY_POINT_VALUE    1
    #elif configUNIQUE_INTERRUPT_PRIORITIES == 128
        #define portPRIORITY_SHIFT            1
        #define portMAX_BINARY_POINT_VALUE    0
    #elif configUNIQUE_INTERRUPT_PRIORITIES == 256
        #define portPRIORITY_SHIFT            0
        #define portMAX_BINARY_POINT_VALUE    0
    #else /* if configUNIQUE_INTERRUPT_PRIORITIES == 16 */
        #error Invalid configUNIQUE_INTERRUPT_PRIORITIES setting.  configUNIQUE_INTERRUPT_PRIORITIES must be set to the number of unique priorities implemented by the target hardware
    #endif /* if configUNIQUE_INTERRUPT_PRIORITIES == 16 */

/* Interrupt controller access addresses. */
    #define portICCPMR_PRIORITY_MASK_OFFSET                      ( 0x04 )
    #define portICCIAR_INTERRUPT_ACKNOWLEDGE_OFFSET              ( 0x0C )
    #define portICCAIAR_INTERRUPT_ACKNOWLEDGE_OFFSET              ( 0x20 )
    #define portICCEOIR_END_OF_INTERRUPT_OFFSET                  ( 0x10 )
    #define portICCAEOIR_END_OF_INTERRUPT_OFFSET                  ( 0x24 )
    #define portICCBPR_BINARY_POINT_OFFSET                       ( 0x08 )
    #define portICCRPR_RUNNING_PRIORITY_OFFSET                   ( 0x14 )

    #define portINTERRUPT_CONTROLLER_CPU_INTERFACE_ADDRESS       ( configINTERRUPT_CONTROLLER_BASE_ADDRESS + configINTERRUPT_CONTROLLER_CPU_INTERFACE_OFFSET )
    #define portICCPMR_PRIORITY_MASK_REGISTER                    ( *( ( volatile uint32_t * ) ( portINTERRUPT_CONTROLLER_CPU_INTERFACE_ADDRESS + portICCPMR_PRIORITY_MASK_OFFSET ) ) )
    #define portICCIAR_INTERRUPT_ACKNOWLEDGE_REGISTER_ADDRESS    ( portINTERRUPT_CONTROLLER_CPU_INTERFACE_ADDRESS + portICCAIAR_INTERRUPT_ACKNOWLEDGE_OFFSET )
    #define portICCEOIR_END_OF_INTERRUPT_REGISTER_ADDRESS        ( portINTERRUPT_CONTROLLER_CPU_INTERFACE_ADDRESS + portICCAEOIR_END_OF_INTERRUPT_OFFSET )
    #define portICCPMR_PRIORITY_MASK_REGISTER_ADDRESS            ( portINTERRUPT_CONTROLLER_CPU_INTERFACE_ADDRESS + portICCPMR_PRIORITY_MASK_OFFSET )
    #define portICCBPR_BINARY_POINT_REGISTER                     ( *( ( const volatile uint32_t * ) ( portINTERRUPT_CONTROLLER_CPU_INTERFACE_ADDRESS + portICCBPR_BINARY_POINT_OFFSET ) ) )
    #define portICCRPR_RUNNING_PRIORITY_REGISTER                 ( *( ( const volatile uint32_t * ) ( portINTERRUPT_CONTROLLER_CPU_INTERFACE_ADDRESS + portICCRPR_RUNNING_PRIORITY_OFFSET ) ) )

    #define portMEMORY_BARRIER()    __asm volatile ( "" ::: "memory" )

typedef struct {
	union {
		uint32_t slock;
		struct __raw_tickets {
#ifdef __ARMEB__
			uint16_t next;
			uint16_t owner;
#else
			uint16_t owner;
			uint16_t next;
#endif
		} tickets;
	};
} arch_spinlock_t;

#define barrier() __asm__ __volatile__("": : :"memory")
#define smp_mb() __asm__ __volatile__("DMB\n")
#define dsb_sev() __asm__ __volatile__("DSB\n			\
										SEV\n")

#define __READ_ONCE_SIZE                        \
({                                  \
    switch (size) {                         \
    case 1: *(uint8_t *)res = *(volatile uint8_t *)p; break;      \
    case 2: *(uint16_t *)res = *(volatile uint16_t *)p; break;        \
    case 4: *(uint32_t *)res = *(volatile uint32_t *)p; break;        \
    case 8: *(uint64_t *)res = *(volatile uint64_t *)p; break;        \
    default:                            \
        barrier();                      \
        memcpy((void *)res, (const void *)p, size);   \
        barrier();                      \
    }                               \
})


#define TICKET_SHIFT	16
#define READ_ONCE(x) __READ_ONCE(x, 1)

#define __READ_ONCE(x, check)                       \
({                                  \
    union { typeof(x) __val; char __c[1]; } __u;            \
    if (check)                          \
        __read_once_size(&(x), __u.__c, sizeof(x));     \
    else                                \
        __read_once_size_nocheck(&(x), __u.__c, sizeof(x)); \
    __u.__val;                          \
})

static inline void __read_once_size(const volatile void *p, void *res, int size)
{
    __READ_ONCE_SIZE;
}

static inline void __read_once_size_nocheck(const volatile void *p, void *res, int size)
{
    __READ_ONCE_SIZE;
}

#define WRITE_ONCE(x, val) \
({                          \
    union { typeof(x) __val; char __c[1]; } __u =   \
        { .__val = (__force typeof(x)) (val) }; \
    __write_once_size(&(x), __u.__c, sizeof(x));    \
    __u.__val;                  \
})

static inline void __write_once_size(volatile void *p, void *res, int size)
{
    switch (size) {
    case 1: *(volatile uint8_t *)p = *(uint8_t *)res; break;
    case 2: *(volatile uint16_t *)p = *(uint16_t *)res; break;
    case 4: *(volatile uint32_t *)p = *(uint32_t *)res; break;
    case 8: *(volatile uint64_t *)p = *(uint64_t *)res; break;
    default:
        barrier();
        memcpy((void *)p, (const void *)res, size);
        barrier();
    }
}


static inline void arch_spin_lock(arch_spinlock_t *lock)
{
    unsigned long tmp;
    uint32_t newval;
    arch_spinlock_t lockval;

    __builtin_prefetch(&lock->slock);
    __asm__ __volatile__(
"1: ldrex   %0, [%3]\n"
"   add %1, %0, %4\n"
"   strex   %2, %1, [%3]\n"
"   teq %2, #0\n"
"   bne 1b"
    : "=&r" (lockval), "=&r" (newval), "=&r" (tmp)
    : "r" (&lock->slock), "I" (1 << TICKET_SHIFT)
    : "cc");

    while (lockval.tickets.next != lockval.tickets.owner) {
        __asm__ __volatile__("wfi");
        lockval.tickets.owner = READ_ONCE(lock->tickets.owner);
    }
    smp_mb();
}

static inline int arch_spin_trylock(arch_spinlock_t *lock)
{
	unsigned long contended, res;
	uint32_t slock;

	__builtin_prefetch(&lock->slock);
	do {
		__asm__ __volatile__(
		"	ldrex	%0, [%3]\n"
		"	mov	%2, #0\n"
		"	subs	%1, %0, %0, ror #16\n"
		"	addeq	%0, %0, %4\n"
		"	strexeq	%2, %0, [%3]"
		: "=&r" (slock), "=&r" (contended), "=&r" (res)
		: "r" (&lock->slock), "I" (1 << TICKET_SHIFT)
		: "cc");
	} while (res);

	if (!contended) {
		smp_mb();
		return 1;
	} else {
		return 0;
	}
}

static inline void arch_spin_unlock(arch_spinlock_t *lock)
{
	smp_mb();
	lock->tickets.owner++;
	dsb_sev();
}

/* Multi-core */
    #define portSUPPORT_SMP 1 /* this is an SMP build which means configNUM_CORES is relevant */
    #define portMAX_CORE_COUNT            2
    #ifndef configNUM_CORES
    #define configNUM_CORES               2
    #endif
    /* FreeRTOS core id is always zero based, so always 0 if we're running on only one core */
#if configNUM_CORES == portMAX_CORE_COUNT
    int get_core_id(void);
    #define portGET_CORE_ID() get_core_id()
#else
    #define portGET_CORE_ID() 0
#endif


    #define portTICK_TYPE_IS_ATOMIC    1

    /* Requires for SMP */
    #define portCRITICAL_NESTING_IN_TCB   1

    void vYieldCore( int xCoreID );
#define portYIELD_CORE(a) vYieldCore(a)


    #define portRTOS_SPINLOCK_COUNT 2

    /* Note this is a single method with uxAcquire parameter since we have
     * static vars, the method is always called with a compile time constant for
     * uxAcquire, and the compiler should dothe right thing! */
    #if 0
    static inline void vPortRecursiveLock(uint32_t ulLockNum, arch_spinlock_t *pxSpinLock, BaseType_t uxAcquire) {
        static uint8_t ucOwnedByCore[ portMAX_CORE_COUNT ];
        static uint8_t ucRecursionCountByLock[ portRTOS_SPINLOCK_COUNT ];
        configASSERT(ulLockNum >= 0 && ulLockNum < portRTOS_SPINLOCK_COUNT );
        uint32_t ulCoreNum = portGET_CORE_ID();
        uint32_t ulLockBit = 1u << ulLockNum;
        uint32_t ulState = 0;
        configASSERT(ulLockBit < 256u );
        if( uxAcquire )
        {
            if( __builtin_expect( (pxSpinLock->slock == ulCoreNum), 0 ) )  //已经占又该锁
            {

                if( ucOwnedByCore[ulCoreNum] & ulLockBit )
                {
                    configASSERT(ucRecursionCountByLock[ulLockNum] != 255u );
                    ucRecursionCountByLock[ulLockNum]++;
                    return;
                }
                while ( __builtin_expect( (pxSpinLock->slock == ulCoreNum), 0 ) );
            }
            //__mem_fence_acquire();
            //pxSpinLock->slock = 0;
            configASSERT(ucRecursionCountByLock[ulLockNum] == 0 );
            ulState = portDISABLE_INTERRUPTS();
            ucRecursionCountByLock[ulLockNum] = 1;
            ucOwnedByCore[ulCoreNum] |= ulLockBit;
            lockMutex(pxSpinLock);
            portRESTORE_INTERRUPTS(ulState);
        } else {
            configASSERT((ucOwnedByCore[ulCoreNum] & ulLockBit) != 0 );
            configASSERT(ucRecursionCountByLock[ulLockNum] != 0 );
            ulState = portDISABLE_INTERRUPTS();
            if( !--ucRecursionCountByLock[ulLockNum] )
            {
                unlockMutex(pxSpinLock);
                ucOwnedByCore[ulCoreNum] &= ~ulLockBit;
                //__mem_fence_release();
                //pxSpinLock->slock = 0xFF;
            }
            portRESTORE_INTERRUPTS(ulState);
        }
    }
    #endif

static inline void vPortRecursiveLock(uint32_t ulLockNum, arch_spinlock_t *pxSpinLock, BaseType_t uxAcquire) 
{
    static uint8_t ucRecursionCountByLock[ portRTOS_SPINLOCK_COUNT ];
    configASSERT(ulLockNum >= 0 && ulLockNum < portRTOS_SPINLOCK_COUNT );
    uint32_t ulCoreNum = portGET_CORE_ID();
    uint32_t ulLockBit = 1u << ulLockNum;
    uint32_t ulState = 0;
    configASSERT(ulLockBit < 256u);
    if( uxAcquire )
    {
        if( __builtin_expect( (pxSpinLock->slock == ulCoreNum), 0 ) )  //该核已经占用该锁+1
        {
            configASSERT(ucRecursionCountByLock[ulLockNum] != 255u );
            ucRecursionCountByLock[ulLockNum]++;
            return;
        }

        ulState = portDISABLE_INTERRUPTS();
        lockMutex(pxSpinLock);          //尝试获取锁
        ucRecursionCountByLock[ulLockNum] = 1;  //获取到锁 计数+1
        portRESTORE_INTERRUPTS(ulState);
    }
    else
    {
        ulState = portDISABLE_INTERRUPTS();

        if(__builtin_expect((pxSpinLock->slock != ulCoreNum), 0))
        {
            //while (1)
            if(pxSpinLock->slock != 0xFF)
            {
                test_putc(0, '@');
            }

            {
                ;
            }
            portRESTORE_INTERRUPTS(ulState);
            return;
        }

        if(ucRecursionCountByLock[ulLockNum] == 0)
        {
            //while (1)
            test_putc(0, '*');
            {
                ;
            }
            unlockMutex(pxSpinLock);
            portRESTORE_INTERRUPTS(ulState);
            return;
        }
        
        //if( __builtin_expect((pxSpinLock->slock == ulCoreNum), 0))  //该核占有该锁，释放锁-1
        {
            configASSERT(ucRecursionCountByLock[ulLockNum] != 0 );
            if(!--ucRecursionCountByLock[ulLockNum])  //该核已经释放所有占用锁调用
            {
                unlockMutex(pxSpinLock); //释放该锁
            }
        }
        portRESTORE_INTERRUPTS(ulState);
    }
}
    extern arch_spinlock_t configSMP_SPINLOCK_0;
    extern arch_spinlock_t configSMP_SPINLOCK_1;

/**
 * @brief 
 * 
 */
#if 0
#define portCHECK_IF_IN_ISR()({                                         \
        uint32_t ulCPSR;                                                \
       __asm__ volatile ("mrs %0, CPSR" : "=r" (ulCPSR)::);             \
       ((((uint8_t)ulCPSR&0x0C)==0) && (((uint8_t)ulCPSR&0x03)!=0));})

#endif

    #define portCHECK_IF_IN_ISR() Check_IfInIsr()
    #define portASSERT_IF_IN_ISR() configASSERT( portCHECK_IF_IN_ISR() == 0 )
    
    //#define portGET_ISR_LOCK()      arch_spin_lock(&configSMP_SPINLOCK_0)    //vPortRecursiveLock(0, spin_lock_instance(configSMP_SPINLOCK_0), pdTRUE)
    //#define portRELEASE_ISR_LOCK()  arch_spin_unlock(&configSMP_SPINLOCK_0)  //vPortRecursiveLock(0, spin_lock_instance(configSMP_SPINLOCK_0), pdFALSE)
    //#define portGET_TASK_LOCK()     arch_spin_lock(&configSMP_SPINLOCK_1)    //vPortRecursiveLock(1, spin_lock_instance(configSMP_SPINLOCK_1), pdTRUE)
    //#define portRELEASE_TASK_LOCK() arch_spin_unlock(&configSMP_SPINLOCK_1)  //vPortRecursiveLock(1, spin_lock_instance(configSMP_SPINLOCK_1), pdFALSE)
    #define portGET_ISR_LOCK()      vPortRecursiveLock(0, &configSMP_SPINLOCK_0, pdTRUE)
    #define portRELEASE_ISR_LOCK()  vPortRecursiveLock(0, &configSMP_SPINLOCK_0, pdFALSE)
    #define portGET_TASK_LOCK()     vPortRecursiveLock(1, &configSMP_SPINLOCK_1, pdTRUE)
    #define portRELEASE_TASK_LOCK() vPortRecursiveLock(1, &configSMP_SPINLOCK_1, pdFALSE)




#endif /* PORTMACRO_H */
